/*
DISPLAY PINGS
 Made by: Lars van der Valk
 14/01/2023
 This code looks for certain data to be recieved and then displays a "ping" on location A or B depending
 on the detection of something on A or on B. It also counts up in the right corner for both total times 
 detected on per point.
 
 This code runs on a 1920x1080 screen correctly.
 */

import processing.serial.*;

Serial myPort;

String myString = null;
PImage bg;  //backgroundimage
int i = 0;  //counter for B
int j = 0;  //counter for A
int countA = 0;  //storage for amount of times detected
int countB = 0;  //storage for amount of times detected

long startMillis;
long currentMillis;
long period = 400;

String locator;    //data recieving
char locator2;     //data recieving

void setup() {
  fullScreen();    //set screen size
  frameRate(60);   //set frame rate
  String portName = Serial.list()[0];    //insert the right port number (0, 1 or 2 in most cases)
  myPort = new Serial(this, portName, 115200);  //setup recieving port at 115200 baud
  bg = loadImage("HeerdePlane.png");      //load image for background
  startMillis = millis();     //start milllis timer
}

void draw() {

  image(bg, 0, 0);  //display image
  fill(200);
  textSize(36);
  stroke(0);
  text("A", 1100-10, 250);   //location marker A
  text("B", 800-10, 400);    //location marker B
  noStroke();
  textSize(64);
  text("A: " + countA, width-250, 100);  //counter A
  text("B: " +countB, width-250, 150);   //counter B

  if ( myPort.available() > 0) {  // If data is available,
    myString = myPort.readString();
    myString = myString.replaceAll("\n", "");  //delete all carriage returns
    if (myString != null) {
      String[] list = split(myString, ",");  //split the string on ,
      locator = list[0];    //look at first string
      locator2 = locator.charAt(0);  //look at first character of first string
    } else {
      locator2 = 0;  //else write as an unused character
    }
  }


  if (i<100) {
    pushMatrix();
    fill(211, 42, 42, abs(255-i*2.55));  //the higher i gets the more 'see through' the ellipse
    translate(800, 400);    //place the marker in the right place on B
    scale (i*i/800);    //increase size of ellipse dependent on i
    ellipse(0, 0, 50, 18);    //draw ellipse
    popMatrix();
    i++;    // count up
    locator2 = 0;  //set locator as unused character to only trigger once
  } else if (locator2 == '2') {  // else look if locater has a detection value
    i = 0;  //set i as zero to run above loop
    countB++;  //count up the counter
  }

  if (j<100) {
    pushMatrix();
    fill(211, 42, 42, abs(255-j*2.55));  //the higher i gets the more 'see through' the ellipse
    translate(1100, 250);    //place the marker in the right place on A
    scale (j*j/800);    //increase size of ellipse dependent on i
    ellipse(0, 0, 50, 18);    //draw ellipse
    popMatrix();
    j++;    // count up
    locator2 = 0; //set locator as unused character to only trigger once
  } else if (locator2 == '1') { // else look if locater has a detection value
    j = 0;  //set j as zero to run above loop
    countA++; //count up the counter
  }
}
