/*
   MASTER
   Made by: Lars van der Valk
   14/01/2023
   This code runs on ESP32 V-room
   It communicates with another ESP to indicate when a boar is spotted and close by
   Connected:
     -Ultrasonic sensor
     -Buzzer/Speaker
     -Infrared 8x8 pixel camera

   Built for the project of module 2 CreaTe
   Based on the ESP-Now example of Arvind Ravulavaru <https://github.com/arvindr21>
   This device works as a master sending information to the 'slave'

   The ultrasonic speaker turns on the IR camera which detects if the thing in front of
   the camera is something warmer than 22 degrees. Then it will send data to the "slave".


         NOTE: These terms are outdated but used for clarity,
         i do not condone the concept of master and slave outside of the bedroom ;)
*/

//__INCLUDING LIBRARIES__//
//Including libraries communication//
#include <esp_now.h>
#include <WiFi.h>
#include <esp_wifi.h> // only for esp_wifi_set_channel()

//Include library thermal imaging
#include <Adafruit_AMG88xx.h>


//__DEFINITIONS__//
#define SOUND_SPEED 0.034   //distance measuring 
#define NOTE  440       //note played when sensed


//Global copy of slave//
esp_now_peer_info_t slave;
#define CHANNEL 1
#define PRINTSCANRESULTS 0
#define DELETEBEFOREPAIR 0


//__DECLARING VARIABLES__//
Adafruit_AMG88xx amg;                    //Initialize thermal sensor AMG8833
float pixels[AMG88xx_PIXEL_ARRAY_SIZE];  //Initialize array pixels
const int echoPin = 27;                  //Distance sensing
const int trigPin = 26;                  //Distance sensing
const int relayPin = 4;                  //Controlling the relay for the LEDs
const int speakerPin = 5;                //Playing note
long duration;                           //Playing note
float currentDistanceCm;                 //Used to calculate distance (data-smoothing)
float newDistanceCm;                     //Used to calculate distance (data-smoothing)
uint8_t data = 0;                        //Initialize data sent over ESPNOW protocol
unsigned long startMillis = 0;               //Start time
unsigned long currentMillis;             //Actual time for loop
const unsigned long period = 300;        //Period for the loop based on start and actual time
unsigned long currentMillis2;            //Actual time for loop
const unsigned long period2 = 3000;      //Period for the IR sensor loop
unsigned long currentMillisL;            //Actual time for loop
unsigned long startMillisL = 0;
unsigned long startMillisD = 0;
const unsigned long ledPeriod = 200;         //Period for the LED on/off
unsigned long currentMillis4;
unsigned long startMillis4;
const unsigned long period4 = 100;
int relayState = LOW;
float boarHeat = 23;
bool boarDetected = false;
bool act = false;

//Initialize ESP Now with fallback//
void InitESPNow() {
  WiFi.disconnect();
  if (esp_now_init() == ESP_OK) {
    //    Serial.println("ESPNow Init Success");
  }
  else {
    //    Serial.println("ESPNow Init Failed");
    // Retry InitESPNow, add a counte and then restart?
    // InitESPNow();
    // or Simply Restart
    ESP.restart();
  }
}

// Scan for slaves in AP mode//
void ScanForSlave() {
  int16_t scanResults = WiFi.scanNetworks(false, false, false, 300, CHANNEL); // Scan only on one channel
  // reset on each scan
  bool slaveFound = 0;
  memset(&slave, 0, sizeof(slave));

  //Serial.println("");
  if (scanResults == 0) {
    //Serial.println("No WiFi devices in AP Mode found");
  } else {
    //Serial.print("Found "); Serial.print(scanResults); Serial.println(" devices ");
    for (int i = 0; i < scanResults; ++i) {
      // Print SSID and RSSI for each device found
      String SSID = WiFi.SSID(i);
      int32_t RSSI = WiFi.RSSI(i);
      String BSSIDstr = WiFi.BSSIDstr(i);

      if (PRINTSCANRESULTS) {
        //        Serial.print(i + 1);
        //        Serial.print(": ");
        //        Serial.print(SSID);
        //        Serial.print(" (");
        //        Serial.print(RSSI);
        //        Serial.print(")");
        //        Serial.println("");
      }
      delay(10);
      // Check if the current device starts with `Slave`
      if (SSID.indexOf("Slave") == 0) {
        // SSID of interest
        //        Serial.println("Found a Slave.");
        //        Serial.print(i + 1); Serial.print(": "); Serial.print(SSID); Serial.print(" ["); Serial.print(BSSIDstr); Serial.print("]"); Serial.print(" ("); Serial.print(RSSI); Serial.print(")"); Serial.println("");
        // Get BSSID => Mac Address of the Slave
        int mac[6];
        if ( 6 == sscanf(BSSIDstr.c_str(), "%x:%x:%x:%x:%x:%x",  &mac[0], &mac[1], &mac[2], &mac[3], &mac[4], &mac[5] ) ) {
          for (int ii = 0; ii < 6; ++ii ) {
            slave.peer_addr[ii] = (uint8_t) mac[ii];
          }
        }

        slave.channel = CHANNEL; // pick a channel
        slave.encrypt = 0; // no encryption

        slaveFound = 1;
        // we are planning to have only one slave in this example;
        // Hence, break after we find one, to be a bit efficient
        break;
      }
    }
  }

  if (slaveFound) {
    //    Serial.println("Slave Found, processing..");
  } else {
    //    Serial.println("Slave Not Found, trying again.");
  }

  // clean up ram
  WiFi.scanDelete();
}

void deletePeer() {
  esp_err_t delStatus = esp_now_del_peer(slave.peer_addr);
  //  Serial.print("Slave Delete Status: ");
  if (delStatus == ESP_OK) {
    // Delete success
    //    Serial.println("Success");
  } else if (delStatus == ESP_ERR_ESPNOW_NOT_INIT) {
    // How did we get so far!!
    //    Serial.println("ESPNOW Not Init");
  } else if (delStatus == ESP_ERR_ESPNOW_ARG) {
    //    Serial.println("Invalid Argument");
  } else if (delStatus == ESP_ERR_ESPNOW_NOT_FOUND) {
    //    Serial.println("Peer not found.");
  } else {
    //    Serial.println("Not sure what happened");
  }
}


// Check if the slave is already paired with the master.
// If not, pair the slave with master
bool manageSlave() {
  if (slave.channel == CHANNEL) {
    if (DELETEBEFOREPAIR) {
      deletePeer();
    }
    //Serial.print("Slave Status: ");
    // check if the peer exists
    bool exists = esp_now_is_peer_exist(slave.peer_addr);
    if ( exists) {
      // Slave already paired.
      //Serial.println("Already Paired");
      return true;
    } else {
      // Slave not paired, attempt pair
      esp_err_t addStatus = esp_now_add_peer(&slave);
      if (addStatus == ESP_OK) {
        // Pair success
        //  Serial.println("Pair success");
        return true;
      } else if (addStatus == ESP_ERR_ESPNOW_NOT_INIT) {
        // How did we get so far!!
        //        Serial.println("ESPNOW Not Init");
        return false;
      } else if (addStatus == ESP_ERR_ESPNOW_ARG) {
        //        Serial.println("Invalid Argument");
        return false;
      } else if (addStatus == ESP_ERR_ESPNOW_FULL) {
        //        Serial.println("Peer list full");
        return false;
      } else if (addStatus == ESP_ERR_ESPNOW_NO_MEM) {
        //        Serial.println("Out of memory");
        return false;
      } else if (addStatus == ESP_ERR_ESPNOW_EXIST) {
        //  Serial.println("Peer Exists");
        return true;
      } else {
        // Serial.println("Not sure what happened");
        return false;
      }
    }
  } else {
    // No slave found to process
    //    Serial.println("No Slave found to process");
    return false;
  }
}

//__SEND DATA__//
void sendData() {
  const uint8_t *peer_addr = slave.peer_addr;
  //  Serial.print("Sending: "); Serial.println(data);
  esp_err_t result = esp_now_send(peer_addr, &data, sizeof(data));
  //  Serial.print("Send Status: ");
  if (result == ESP_OK) {
    //    Serial.println("Success");
  } else if (result == ESP_ERR_ESPNOW_NOT_INIT) {
    // How did we get so far!!
    //    Serial.println("ESPNOW not Init.");
  } else if (result == ESP_ERR_ESPNOW_ARG) {
    //    Serial.println("Invalid Argument");
  } else if (result == ESP_ERR_ESPNOW_INTERNAL) {
    //    Serial.println("Internal Error");
  } else if (result == ESP_ERR_ESPNOW_NO_MEM) {
    //    Serial.println("ESP_ERR_ESPNOW_NO_MEM");
  } else if (result == ESP_ERR_ESPNOW_NOT_FOUND) {
    //    Serial.println("Peer not found.");
  } else {
    //    Serial.println("Not sure what happened");
  }
}

// callback when data is sent from Master to Slave
void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status) {
  char macStr[18];
  snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
           mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5]);
  //  Serial.print("Last Packet Sent to: "); Serial.println(macStr);
  //  Serial.print("Last Packet Send Status: "); Serial.println(status == ESP_NOW_SEND_SUCCESS ? "Delivery Success" : "Delivery Fail");
}


//__VOID SETUP__//

void setup() {
  Serial.begin(115200); //set baud rate
  WiFi.mode(WIFI_STA);  //Set device in STA mode to begin with
  delay(200);

  esp_wifi_set_channel(CHANNEL, WIFI_SECOND_CHAN_NONE);

  pinMode(relayPin, OUTPUT);    //Set relay control as output
  digitalWrite(relayPin, relayState);  //Set relay to low

  //Print macAdress and channel when initializing
  //  Serial.print("STA MAC: "); Serial.println(WiFi.macAddress());
  //  Serial.print("STA CHANNEL "); Serial.println(WiFi.channel());

  //Fallback logic
  bool status;
  // default settings
  status = amg.begin();
  if (!status) {
    //    Serial.println("Could not find a valid AMG88xx sensor, check wiring!");
    while (1);
  }

  // Init ESPNow with a fallback logic
  InitESPNow();
  // Once ESPNow is successfully Initialized, we will register for Send CB to
  esp_now_register_send_cb(OnDataSent);   // get the status of Trasnmitted packet

  //__SENSING__//
  //Ultrasonic distance sensor
  pinMode(trigPin, OUTPUT);   // Sets the trigPin as an Output
  pinMode(echoPin, INPUT);    // Sets the echoPin as an Input

  delay(100);               //Lt the sensors boot up
  startMillis = millis();   //Start a millis timer for timing in the code
  startMillisL = millis();
  startMillis4 = millis();
  startMillisD = millis();

}

//__VOID LOOP__//

void loop() {
  //Check if a boar is present
  currentMillis4 = millis();
  if (currentMillis4 - startMillis4 >= period4) {
    if (newDistanceCm < 34) {
      Serial.println("Detect under 34");
      float sum = 0;
      float average;
      amg.readPixels(pixels);  //read all the pixels
      for (int i = 0; i < AMG88xx_PIXEL_ARRAY_SIZE; i++) {
        sum += pixels[i];
      }
      average = sum / AMG88xx_PIXEL_ARRAY_SIZE;
      Serial.println(average);
      if (average >= boarHeat) {
        boarDetected = true;
        //Serial.println("BOAR DETECTED");
      } else {
        boarDetected = false;
      }
    } else {
      boarDetected = false;
    }
    startMillis4 = currentMillis4;
  }


  //__WORK THE ULTRASONIC DISTANCE SENSOR__//
  digitalWrite(trigPin, LOW); // Clears the trigPin
  delayMicroseconds(2);
  digitalWrite(trigPin, HIGH); // Sets the trigPin on HIGH state for 10 micro seconds
  delayMicroseconds(10);
  digitalWrite(trigPin, LOW);
  // Reads the echoPin, returns the sound wave travel time in microseconds
  duration = pulseIn(echoPin, HIGH);
  // Calculate the distance
  currentDistanceCm = duration * SOUND_SPEED / 2;
  //smooth the distance reading by only passing through
  //changes which are larger than half a centimter
  if (abs(currentDistanceCm - newDistanceCm) > 0.5) {
    newDistanceCm = currentDistanceCm;

  }

  if (boarDetected) {
    blinkLights();  // call the void to blink lights
    currentMillis2 = millis();
    if (currentMillis2 - startMillisD >= period2) {
      tone(speakerPin, NOTE);     //play a note for a certain period
      startMillisD = currentMillis2;    
    }
    data = 1;
  } else {
    noTone(speakerPin);
    digitalWrite(relayPin, LOW);
    data = 0;
  }

  if (data == 1) {
    currentMillis = millis();
    if (currentMillis - startMillis >= period) {
      // In the loop we scan for slave
      ScanForSlave();
      // If Slave is found, it would be populate in `slave` variable
      // We will check if `slave` is defined and then we proceed further
      if (slave.channel == CHANNEL) { // check if slave channel is defined
        // `slave` is defined
        // Add slave as peer if it has not been added already
        bool isPaired = manageSlave();
        if (isPaired) {
          // pair success or already paired
          // Send data to device
          sendData();
        } else {
          // slave pair failed
          //        Serial.println("Slave pair failed!");
        }
      }
      else {
        // No slave found to process
      }
      startMillis = currentMillis;    //Reset the counter to recalculate a given period passing
    }
  }
}

//__VOID BLINK LIGHTS__//
void blinkLights() {
  Serial.println("blinks");
  for (int i = 0; i < 5; i++) {
    currentMillisL = millis();
    if (currentMillisL - startMillisL >= ledPeriod) {
      if (relayState == LOW) {
        relayState = HIGH;
      }
      else {
        relayState = LOW;
      }
      digitalWrite (relayPin, relayState);
      startMillisL = currentMillisL;
    }
  }
}
