/*
   SLAVE
   Made by: Lars van der Valk
   14/01/2023
   This code runs on ESP32 V-room
   It communicates with another ESP to indicate when a boar is spotted and close by
   Connected:
     -Ultrasonic sensor
     -1 red LED

   Built for the project of module 2 CreaTe
   Based on the ESP-Now example of Arvind Ravulavaru <https://github.com/arvindr21>
   This device works as a 'slave' getting information from the 'master', combining this
   with its own information and sending that to the laptop. On this a processing code runs.

   The ultrasonic speaker turns on the IR camera which detects if the thing in front of
   the camera is something warmer than 22 degrees. Then it will send data to the "slave".


         NOTE: These terms are outdated but used for clarity,
         i do not condone the concept of master and slave outside of the bedroom ;)
*/

#include <esp_now.h>
#include <WiFi.h>
#define SOUND_SPEED 0.034      //distance measuring 
#define CHANNEL 1

const int echoPin = 26;                  //Distance sensing
const int trigPin = 27;                  //Distance sensing
float currentDistanceCm;                 //Used to calculate distance (data-smoothing)
float newDistanceCm;                     //Used to calculate distance (data-smoothing)
float boarDetectDistanceCm = 15;         //Distance to detect boars in front of the forrest
const int ledPin = 32;                   //Define the relay pin
int ledState = LOW;                      //Define relay State
long duration;                           //Playing note
bool boarDetected = false;

unsigned long currentMillis;
unsigned long currentMillis2;
unsigned long startMillis;
unsigned long startMillis2;

const unsigned long period = 200;
const unsigned long period2 = 200;

// Init ESP Now with fallback
void InitESPNow() {
  WiFi.disconnect();
  if (esp_now_init() == ESP_OK) {
    Serial.println("ESPNow Init Success");
  }
  else {
    Serial.println("ESPNow Init Failed");
    // Retry InitESPNow, add a counte and then restart?
    // InitESPNow();
    // or Simply Restart
    ESP.restart();
  }
}

// config AP SSID
void configDeviceAP() {
  const char *SSID = "Slave_1";
  bool result = WiFi.softAP(SSID, "Slave_1_Password", CHANNEL, 0);
  if (!result) {
    Serial.println("AP Config failed.");
  } else {
    Serial.println("AP Config Success. Broadcasting with AP: " + String(SSID));
    Serial.print("AP CHANNEL "); Serial.println(WiFi.channel());
  }
}

void setup() {
  Serial.begin(115200);
  Serial.println("ESPNow/Basic/Slave Example");
  //Set device in AP mode to begin with
  WiFi.mode(WIFI_AP);
  // configure device AP mode
  configDeviceAP();
  // This is the mac address of the Slave in AP Mode
  Serial.print("AP MAC: "); Serial.println(WiFi.softAPmacAddress());
  // Init ESPNow with a fallback logic
  InitESPNow();
  // Once ESPNow is successfully Init, we will register for recv CB to
  // get recv packer info.
  esp_now_register_recv_cb(OnDataRecv);

  pinMode(ledPin, OUTPUT);           //Set relay control as output
  digitalWrite(ledPin, ledState);  //Set relay to low

  //Ultrasonic distance sensor
  pinMode(trigPin, OUTPUT);   // Sets the trigPin as an Output
  pinMode(echoPin, INPUT);    // Sets the echoPin as an Input

  delay(100);


  startMillis = millis();
  startMillis2 = millis();
}

// callback when data is recv from Master
void OnDataRecv(const uint8_t *mac_addr, const uint8_t *data, int data_len) {
  char macStr[18];
  snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
           mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5]);
  //Serial.print("Last Packet Recv from: "); Serial.println(macStr);
  //Serial.print("Last Packet Recv Data: ");
  Serial.print("2,");
  Serial.println(*data);
}

void loop() {
  currentMillis2 = millis();
  if (currentMillis2 - startMillis2 >= period2) {
    if (newDistanceCm < boarDetectDistanceCm) {
      boarDetected = true;
    } else {
      boarDetected = false;
    }
    if (boarDetected) {
      Serial.print("1,");
      Serial.println(1);
      blinkLED();
      startMillis2 = currentMillis2;
    } else {
      digitalWrite(ledPin, LOW);
    }
    startMillis2 = currentMillis2;
  }


  //__WORK THE ULTRASONIC DISTANCE SENSOR__//
  digitalWrite(trigPin, LOW); // Clears the trigPin
  delayMicroseconds(2);
  digitalWrite(trigPin, HIGH); // Sets the trigPin on HIGH state for 10 micro seconds
  delayMicroseconds(10);
  digitalWrite(trigPin, LOW);
  // Reads the echoPin, returns the sound wave travel time in microseconds
  duration = pulseIn(echoPin, HIGH);
  // Calculate the distance
  currentDistanceCm = duration * SOUND_SPEED / 2;
  //Serial.println(currentDistanceCm);
  //smooth the distance reading by only passing through
  //changes which are larger than half a centimter
  if (abs(currentDistanceCm - newDistanceCm) > 0.5 && currentDistanceCm > 1.2) {
    newDistanceCm = currentDistanceCm;
  }
}

//__BLINK THE LEDS__//
void blinkLED() {
  for (int i = 0; i < 5; i++) {
    currentMillis = millis();
    if (currentMillis - startMillis >= period) {
      if (ledState == LOW) {
        ledState = HIGH;
      }
      else {
        ledState = LOW;
      }
      digitalWrite (ledPin, ledState);
      startMillis = currentMillis;
    }
  }
}
